{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\n# Profiling of dictionary\n\n.. index:: event, profiling, dictionary\n\nImplementation Python dictionary is quite efficient.\nIt is possible to replace it with a C++ implementation?\nThe following code compares several implementation of\na dictionary ``{ str, int: int }``.\n\n## Implementation to compare\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import matplotlib.pyplot as plt\nimport pandas\nfrom cpyquickhelper.fastdata.fast_dict import (\n    FastDictStrInt64, FastDictInt64Int64, FastDictIntInt)\nfrom cpyquickhelper.fastdata.fast_dict_c import (\n    UnorderedMapStrInt64_Create, UnorderedMapStrInt64_Insert,\n    UnorderedMapStrInt64_Get, UnorderedMapStrInt64_Get_Fast,\n    UnorderedMapStrInt64_Insert_Fast,\n    UnorderedMapRHStrInt64_Create, UnorderedMapRHStrInt64_Insert_Fast,\n    UnorderedMapRHStrInt64_Get_Fast)\nfrom pyquickhelper.pycode.profiling import profile, profile2graph\n\n\ndata = [(\"*\" * (i // 2), i, i) for i in range(0, 1000)]\nfcts = []\n\n\nclass UnorderedMapRHStrInt64:\n\n    def __init__(self):\n        self.obj = UnorderedMapRHStrInt64_Create()\n\n    def insert_fast(self, name, key, value):\n        UnorderedMapRHStrInt64_Insert_Fast(self.obj, name, key, value)\n\n    def get_fast(self, name, key, default_value):\n        return UnorderedMapRHStrInt64_Get_Fast(self.obj, name, key, default_value)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Python implementation\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def insert_py(d):\n    for n, k, v in data:\n        d[n, k] = v\n\n\ndef insert_py2(d):\n    for n, k, v in data:\n        if k in d:\n            d[k][n] = v\n        else:\n            d[k] = {n: v}\n\n\ndef fct_py():\n    d1 = {}\n    insert_py(d1)\n    get_py(d1)\n\n\nfcts.append(fct_py)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Second Python implementation\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def g_py(d, n, k, v):\n    return d.get((n, k), 0)\n\n\ndef get_py(d):\n    for n, k, v in data:\n        g_py(d, n, k, v)\n\n\ndef g_py2(d, n, k, v):\n    if k in d:\n        if n in d[k]:\n            return d[k][n]\n    return v\n\n\ndef get_py2(d):\n    for n, k, v in data:\n        g_py2(d, n, k, v)\n\n\ndef fct_py2():\n    d11 = {}\n    insert_py2(d11)\n    get_py2(d11)\n\n\nfcts.append(fct_py2)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "pybind11 + unordered_map\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def insert_c11(d):\n    for n, k, v in data:\n        d.insert(n, k, v)\n\n\ndef get_c11(d):\n    for n, k, _ in data:\n        d.get(n, k, 0)\n\n\ndef fct_c11():\n    d2 = FastDictStrInt64()\n    insert_c11(d2)\n    get_c11(d2)\n\n\nfcts.append(fct_c11)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "C API + unordered_map + wrapper\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class UnorderedMapStrInt64:\n\n    def __init__(self):\n        self.obj = UnorderedMapStrInt64_Create()\n\n    def insert(self, name, key, value):\n        UnorderedMapStrInt64_Insert(self.obj, name, key, value)\n\n    def insert_fast(self, name, key, value):\n        UnorderedMapStrInt64_Insert_Fast(self.obj, name, key, value)\n\n    def get(self, name, key, default_value):\n        return UnorderedMapStrInt64_Get(self.obj, name, key, default_value)\n\n    def get_fast(self, name, key, default_value):\n        return UnorderedMapStrInt64_Get_Fast(self.obj, name, key, default_value)\n\n\ndef insert_c(d):\n    for n, k, v in data:\n        d.insert(n, k, v)\n\n\ndef get_c(d):\n    for n, k, _ in data:\n        d.get(n, k, 0)\n\n\ndef fct_c():\n    d3 = UnorderedMapStrInt64()\n    insert_c(d3)\n    get_c(d3)\n\n\nfcts.append(fct_c)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "C API + unordered_map\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def insert_c2(d):\n    for n, k, v in data:\n        UnorderedMapStrInt64_Insert(d, n, k, v)\n\n\ndef get_c2(d):\n    for n, k, _ in data:\n        UnorderedMapStrInt64_Get(d, n, k, 0)\n\n\ndef fct_c2():\n    d3c = UnorderedMapStrInt64_Create()\n    insert_c2(d3c)\n    get_c2(d3c)\n\n\nfcts.append(fct_c2)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "C API + unordered_map + FAST CALL\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def insert_c_fast(d):\n    for n, k, v in data:\n        UnorderedMapStrInt64_Insert_Fast(d, n, k, v)\n\n\ndef get_c_fast(d):\n    for n, k, _ in data:\n        UnorderedMapStrInt64_Get_Fast(d, n, k, 0)\n\n\ndef fct_c_fast():\n    d3cf = UnorderedMapStrInt64_Create()\n    insert_c_fast(d3cf)\n    get_c_fast(d3cf)\n\n\nfcts.append(fct_c_fast)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "C API + unordered_map + FAST CALL + a different\nimplementation, [martinus/robin-hood-hashing](https://github.com/martinus/robin-hood-hashing)\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def insert_c_fast_rh(d):\n    for n, k, v in data:\n        UnorderedMapRHStrInt64_Insert_Fast(d, n, k, v)\n\n\ndef get_c_fast_rh(d):\n    for n, k, _ in data:\n        UnorderedMapRHStrInt64_Get_Fast(d, n, k, 0)\n\n\ndef fct_c_fast_rh():\n    d3cfrh = UnorderedMapRHStrInt64_Create()\n    insert_c_fast_rh(d3cfrh)\n    get_c_fast_rh(d3cfrh)\n\n\nfcts.append(fct_c_fast_rh)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Profiling\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def fctN(N=4000):\n    for _ in range(N):\n        for f in fcts:\n            f()\n\n\ndef clean_name(text):\n    text = text.replace(\"\\\\\", \"/\")\n    return text.split(\"/examples/\")[-1]\n\n\nps = profile(fctN)[0]\nroot, nodes = profile2graph(ps, clean_text=clean_name)\ntext = root.to_text()\nprint(text)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Visually.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "data = []\nfor node in root:\n    data.append(dict(name=node.func_name, time=node.tall))\ndf = pandas.DataFrame(data).set_index('name').sort_values('time')\n\nfig, ax = plt.subplots(1, 1, figsize=(20, 4))\ndf.plot.barh(title=\"Profiling\", ax=ax)\nfig.tight_layout()\n\n# plt.show()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.9.1"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}